/*!
 *  \file    ucglib_xmega_hal.c
 *  \author  Wim Dolman (<a href="mailto:w.e.dolman@hva.nl">w.e.dolman@hva.nl</a>)
 *  \date    30-05-2020
 *  \version 3.0
 *
 *  \brief   Xmega Hardware Abstraction Layer for ucglib from Oli Kraus
 *
 *  \details This Hardware Abstraction Layer is created confirm the
 *           <a href"https://github.com/olikraus/ucglib/wiki/hal">instructions</a> of Oli Kraus.
 *
 *           The function <code>ucg_connectXmega()</code> sets the connections between the
 *           Xmega and the display. The function <code>ucg_printXmegaConnection()</code> 
 *           prints the connections to the standard output.
 *
 *           This file contains a callback function <code>ucg_com_xmega_bb()</code> that handles the
 *           communication with the Xmega using SPI or bit banging.
 *           The Arduino/C++ implementation of ucglib contains extra printing facilities.
 *           This file contains a bunch of functions that implements the same facilities.
 *           So you can use <code>ucg_SetPrintPos()</code>, <code>ucg_SetPrintDir()</code> 
 *           and in stead of <code>print</code> and <code>println</code> 
 *           you can use <code>ucg_Print()</code> which prints a formatstring.
 *
 *           This print facility uses a hook in the struct ucg_t. Add to the struct these
 *           lines:
 *  \verbatim
 *            #ifdef __AVR_XMEGA__
 *            void  *xmega_hook;   // added pointer for print hook
 *            #endif \endverbatim
 *           or uses the ucg.h of the HAL package in stead. The adapted ucg.h is included  
 *           in this headerfile.
 *
 *           The function ucg_BitmapPrint() print a bitmap to the display. The bitmap
 *           is a headerfile in plain C. The maximum bitmap size is 32768 kB. To show
 *           large images the image must be split in smaller bitmaps.
 */
#ifndef F_CPU
#define F_CPU 320000000UL                 // System clock is 32 MHz
#endif

#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "ucglib_xmega.h"

/*
 * Struct for compatibility printing facilities with Arduino/C++ version of library 
 */
typedef struct {      
  FILE      *fp;      // file pointer for printing 
  ucg_int_t  tx;      // current x coordinate of the position
  ucg_int_t  ty;      // current y coordinate of the position
  uint8_t    tdir;    // current printing direction
} ucg_print_t;           

/*
 * Struct for connection data
 */
typedef struct ucg_xmega_comm_struct {
  void  (*pTransfer) (uint8_t);  
  SPI_t  *pSPI;
  PORT_t *pSCK;
  PORT_t *pSDI; 
  PORT_t *pSDA;
  PORT_t *pCS;
  PORT_t *pRST;
  PORT_t *pCD;
  PORT_t *pBLK;
  uint8_t bpSCK;
  uint8_t bpSDI; 
  uint8_t bpSDA;
  uint8_t bpCS;
  uint8_t bpRST;
  uint8_t bpCD;
  uint8_t bpBLK;
  uint8_t bmSCK;
  uint8_t bmSDI;
  uint8_t bmSDA;
  uint8_t bmCS;
  uint8_t bmRST;
  uint8_t bmCD;
  uint8_t bmBLK;
  uint8_t blkDisabled;
} ucg_xmega_comm_t;

static ucg_t            *_curr_ucg;       // pointer necessary for the printing facilities 
static ucg_xmega_comm_t  _commInterface;  // local struct for connection with Xmega
static pin_t             _pinArray[6];    // local array with six pinconnections

// local functions
static void  _xmega_init(void);
static void  _xmega_transfer_bb(uint8_t data);
static void  _xmega_transfer_spi(uint8_t data);
static void  _xmega_disable(void);
static void  _ucg_print_init(ucg_t *ucg);
static int   _ucg_putc(char c, FILE *stream);
static char  _get_port(PORT_t *p);
static char  _get_spi(SPI_t *s);
static void  _print_port(char c);
static void  _print_bp(uint8_t bp);
static void  _print_bm(uint8_t bm);

// connection display Xmega

/*! \brief  Defines the connections of the display with the Xmega
 *
 *  \param  pInterface   pointer to the SPI-interface
 *  \param  pArray       pointer to an array with pinconnections
 *  \param  blkDisabled  index to disable BLK-connection
 * 
 *          1. If pInterface is NULL bit banging is used.
 *
 *          2. The pointer pArray points to an array with pins.
 *          A pin consist of three elements: an index, a pointer
 *          to a port and a pinposition. 
 *          The last element is UCG_XMEGA_PIN_NULL.
 *          Example 1 using a SPI-interface:
 * \verbatim   pin_t connectArraySPI[] = {
 *                 { UCG_XMEGA_PIN_RST, &PORTD, PIN3_bp },   // RST
 *                 { UCG_XMEGA_PIN_CD,  &PORTD, PIN2_bp },   // CD
 *                 { UCG_XMEGA_PIN_BLK, &PORTD, PIN1_bp },   // BLK
 *                 { UCG_XMEGA_PIN_NULL }
 *             }; \endverbatim
 *
 *          Example 2 using bit banging:
 * \verbatim  pin_t connectArrayBBcompatibelSPID[] = {
 *               { UCG_XMEGA_PIN_SCK, &PORTD, PIN7_bp },   // SCK
 *               { UCG_XMEGA_PIN_SDA, &PORTD, PIN5_bp },   // SDA
 *               { UCG_XMEGA_PIN_CS,  &PORTD, PIN4_bp },   // CS
 *               { UCG_XMEGA_PIN_RST, &PORTD, PIN3_bp },   // RST
 *               { UCG_XMEGA_PIN_CD,  &PORTD, PIN2_bp },   // CD
 *               { UCG_XMEGA_PIN_BLK, &PORTD, PIN1_bp },   // BLK
 *               { UCG_XMEGA_PIN_NULL }    
 *             }; \endverbatim
 *
 *          3. If blkDisabled is 1 the pin of the XMega that is 
 *          connected to BLK will disabled.
 *          On a breadboard it is now possible to connect VCC to BLK.
 *          Be carefull with a powerline of 5V! 
 * 
 *  \return void
 */
void ucg_connectXmega(void *pInterface, pin_t *pArray, uint8_t blkDisabled)
{
  PORT_t *p = NULL;

  // init _pinArray
  for(int i=0; i<6; i++) {
    _pinArray[i].index = 12;
    _pinArray[i].port  = NULL;
    _pinArray[i].bp    = 8;
  }

  // sort pinArray
  while ( pArray->index != 0xFF ) {
    if ( (pArray->index < UCG_XMEGA_PIN_SCK) || (pArray->index > UCG_XMEGA_PIN_BLK) ) continue;
    _pinArray[pArray->index].port = pArray->port;
    _pinArray[pArray->index].bp   = pArray->bp;
    pArray++;
  }

  if (pInterface == NULL) {
    _commInterface.pTransfer = _xmega_transfer_bb;
    _commInterface.pSPI  = NULL;
    _commInterface.pSCK  = _pinArray[UCG_XMEGA_PIN_SCK].port;
    _commInterface.pSDI  = NULL;
    _commInterface.pSDA  = _pinArray[UCG_XMEGA_PIN_SDA].port;
    _commInterface.pCS   = _pinArray[UCG_XMEGA_PIN_CS].port;
    _commInterface.pRST  = _pinArray[UCG_XMEGA_PIN_RST].port;
    _commInterface.pCD   = _pinArray[UCG_XMEGA_PIN_CD].port;
    _commInterface.pBLK  = _pinArray[UCG_XMEGA_PIN_BLK].port;
    _commInterface.bpSCK = _pinArray[UCG_XMEGA_PIN_SCK].bp;
    _commInterface.bpSDI = 8;
    _commInterface.bpSDA = _pinArray[UCG_XMEGA_PIN_SDA].bp;
    _commInterface.bpCS  = _pinArray[UCG_XMEGA_PIN_CS].bp;
    _commInterface.bpRST = _pinArray[UCG_XMEGA_PIN_RST].bp;
    _commInterface.bpCD  = _pinArray[UCG_XMEGA_PIN_CD].bp;
    _commInterface.bpBLK = _pinArray[UCG_XMEGA_PIN_BLK].bp;
  } else {
    _commInterface.pTransfer = _xmega_transfer_spi;
    _commInterface.pSPI  = (SPI_t *) pInterface;
    switch ( (uint16_t) pInterface ) {
      #ifdef SPIC 
        case (uint16_t)&SPIC: p = &PORTC; break;
      #endif
      #ifdef SPID 
        case (uint16_t)&SPID: p = &PORTD; break;
      #endif
      #ifdef SPIE 
        case (uint16_t)&SPIE: p = &PORTE; break;
      #endif
      #ifdef SPIF 
        case (uint16_t)&SPIF: p = &PORTF; break;
      #endif
      #ifdef SPIG 
        case (uint16_t)&SPIG: p = &PORTG; break;
      #endif
    }
    _commInterface.pSCK  = p;
    _commInterface.pSDI  = p;
    _commInterface.pSDA  = p;
    _commInterface.pCS   = p;
    _commInterface.pRST  = _pinArray[UCG_XMEGA_PIN_RST].port;
    _commInterface.pCD   = _pinArray[UCG_XMEGA_PIN_CD].port;
    _commInterface.pBLK  = _pinArray[UCG_XMEGA_PIN_BLK].port;
    _commInterface.bpSCK = PIN7_bp;
    _commInterface.bpSDI = PIN6_bp;
    _commInterface.bmSDI = PIN6_bm;
    _commInterface.bpSDA = PIN5_bp;
    _commInterface.bpCS  = PIN4_bp;
    _commInterface.bpRST = _pinArray[UCG_XMEGA_PIN_RST].bp;
    _commInterface.bpCD  = _pinArray[UCG_XMEGA_PIN_CD].bp;
    _commInterface.bpBLK = _pinArray[UCG_XMEGA_PIN_BLK].bp;
  }
  
  _commInterface.bmSCK = ( 1 << _commInterface.bpSCK );
  _commInterface.bmSDA = ( 1 << _commInterface.bpSDA );
  _commInterface.bmCS  = ( 1 << _commInterface.bpCS  );
  _commInterface.bmRST = ( 1 << _commInterface.bpRST );
  _commInterface.bmCD  = ( 1 << _commInterface.bpCD  );
  _commInterface.bmBLK = ( 1 << _commInterface.bpBLK );

  _commInterface.blkDisabled = blkDisabled;
}


/*! \brief  Print (debugging) the connections with the Xmega
 *
 *  \return void
 */
void ucg_printXmegaConnection(void)
{
  printf("\nConnections Overview\n");
  if (_commInterface.pSPI == NULL) {
    printf("interface     : bit banging\n");
  } else {
    printf("interface     : SPI%c\n", _get_spi(_commInterface.pSPI));
  }
  printf("              : CLK    SDI    SDA    CS     RST    CD     BLK\n");
  printf("port          : ");
  _print_port(_get_port(_commInterface.pSCK));
  _print_port(_get_port(_commInterface.pSDI));
  _print_port(_get_port(_commInterface.pSDA));
  _print_port(_get_port(_commInterface.pCS));
  _print_port(_get_port(_commInterface.pRST));
  _print_port(_get_port(_commInterface.pCD));
  _print_port(_get_port(_commInterface.pBLK));
  printf("\n");

  printf("bit position  : ");
  _print_bp(_commInterface.bpSCK);
  _print_bp(_commInterface.bpSDI);
  _print_bp(_commInterface.bpSDA);
  _print_bp(_commInterface.bpCS);
  _print_bp(_commInterface.bpRST);
  _print_bp(_commInterface.bpCD);
  _print_bp(_commInterface.bpBLK);
  printf("\n");
  printf("bit mask      : ");
  _print_bm(_commInterface.bmSCK);
  _print_bm(_commInterface.bmSDI);
  _print_bm(_commInterface.bmSDA);
  _print_bm(_commInterface.bmCS);
  _print_bm(_commInterface.bmRST);
  _print_bm(_commInterface.bmCD);
  _print_bm(_commInterface.bmBLK);
  printf("\n");
  if ( _commInterface.bpBLK > 7 ) {
    printf("blk           : not used\n");
  } else {
    if ( _commInterface.blkDisabled ) {
      printf("blk           : defined but disabled\n");
    } else {
      printf("blk           : active\n");
    }
  } 
}


// communication

/*! \brief  The callback function for communication between the Xmega and the display.
 *
 *  \param  ucg      pointer to struct for the display
 *  \param  msg      number of the message (action to be done) 
 *  \param  arg      depends on msg: number of arguments, number of microseconds, ...
 *  \param  data     pointer to 8-bit data-array with bytes that needs to be send
 *
 *  \return 16-bit value, always 1
 */
int16_t ucg_comm_xmega(ucg_t *ucg, int16_t msg, uint16_t arg, uint8_t *data)
{
  switch(msg)
  {
    case UCG_COM_MSG_POWER_UP:
      _xmega_init();
      _ucg_print_init(ucg);
      break;
    case UCG_COM_MSG_POWER_DOWN:
      _xmega_disable();
      break;
    case UCG_COM_MSG_DELAY:
      for(uint16_t i=0; i<arg; i++) {
        _delay_us(1);
      }
      break;
    case UCG_COM_MSG_CHANGE_RESET_LINE:
      if (arg) {
        _commInterface.pRST->OUTSET = _commInterface.bmRST;
      } else {
        _commInterface.pRST->OUTCLR = _commInterface.bmRST;
      }
      break;
    case UCG_COM_MSG_CHANGE_CS_LINE:
      if (arg) {
        _commInterface.pCS->OUTSET = _commInterface.bmCS;
      } else {
        _commInterface.pCS->OUTCLR = _commInterface.bmCS;
      }
      break;
    case UCG_COM_MSG_CHANGE_CD_LINE:
      if (arg) {
        _commInterface.pCD->OUTSET = _commInterface.bmCD;
      } else {
        _commInterface.pCD->OUTCLR = _commInterface.bmCD;
      }
      break;
    case UCG_COM_MSG_SEND_BYTE:
      _commInterface.pTransfer(arg);
      break;
    case UCG_COM_MSG_REPEAT_1_BYTE:
      while( arg > 0 ) {
        _commInterface.pTransfer(data[0]);
        arg--;
      }
      break;
    case UCG_COM_MSG_REPEAT_2_BYTES:
      while( arg > 0 ) {
        _commInterface.pTransfer(data[0]);
        _commInterface.pTransfer(data[1]);
        arg--;
      }
      break;
    case UCG_COM_MSG_REPEAT_3_BYTES:
      while( arg > 0 ) {
        _commInterface.pTransfer(data[0]);
        _commInterface.pTransfer(data[1]);
        _commInterface.pTransfer(data[2]);
        arg--;
      }
      break;
    case UCG_COM_MSG_SEND_STR:
      while( arg > 0 ) {
        _commInterface.pTransfer(*data++);
        arg--;
        }
      break;
  }
  
  return 1;
}


// printing facilities

/*! \brief  Gets the current position of the 'print cursor'
 *
 *  \param  ucg      pointer to struct for the display
 *  \param  x        pointer to a variable for the x-coordinate of the current position
 *  \param  y        pointer to a variable for the y-coordinate of the current position
 *
 *  \return void
 */
void ucg_GetPrintPos(ucg_t *ucg, ucg_int_t *x, ucg_int_t *y)
{
  *x = ((ucg_print_t *)ucg->xmega_hook)->tx;
  *y = ((ucg_print_t *)ucg->xmega_hook)->ty;
}


/*! \brief  Sets the position for next "print" command.
 *
 *  \param  ucg      pointer to struct for the display
 *  \param  x        x-coordinate of the position
 *  \param  y        y-coordinate of the position
 *
 *  \return void
 */
void ucg_SetPrintPos(ucg_t *ucg, ucg_int_t x, ucg_int_t y)
{
  ((ucg_print_t *)ucg->xmega_hook)->tx = x;
  ((ucg_print_t *)ucg->xmega_hook)->ty = y;
}


/*! \brief  Sets the direction for next "print" command.
 *
 *  \param  ucg      pointer to struct for the display
 *  \param  dir      the direction
 *
 *  \return void
 */
void ucg_SetPrintDir(ucg_t *ucg, uint8_t dir)
{
  ((ucg_print_t *)ucg->xmega_hook)->tdir = dir;
}


/*! \brief  Put a formatted string to the display at the
 *          current position and in the current direction.
 *
 *          This replaces print and println from the Arduino implementation of ucg_lib
 *
 *          The Arduino style:
 * \verbatim
 *          ucg.print("text ");
 *          ucg.print(x);     // x is an int
 *          ucg.print(" more text ");
 *          ucg.print(y);     // y is a float
 *          ucg.println(";"); \endverbatim
 *              
 *          The replacement in Xmega style:
 *\verbatim
 *          ucg_Print(&ucg, "text %d more text %f;\n", x, y); \endverbatim
 *
 *  \param  ucg      pointer to struct for the display
 *  \param  fmt      formatstring with escape sequences
 *  \param  ...      variables that are printed
 *
 *  \return void
 */
void ucg_Print(ucg_t *ucg, char *fmt, ...)
{
  va_list vl;

  _curr_ucg = ucg;
  va_start(vl, fmt);
  vfprintf( ((ucg_print_t *) (_curr_ucg->xmega_hook))->fp, fmt, vl);
  va_end(vl);
}

/*  brief  Initializes the printing facilities compatible with Arduino/C++ version of library
 *
 *  param  ucg      pointer to struct for the display
 *
 *  return void
 */
static void _ucg_print_init(ucg_t *ucg)
{
  if  (ucg->xmega_hook != NULL) return;
  
  ucg_print_t *p = (ucg_print_t *) malloc(sizeof(ucg_print_t));
  
  if (p == NULL) return;
  
  p->fp    = fdevopen(_ucg_putc, NULL); 
  p->tx    = 0;
  p->ty    = 0;
  p->tdir  = 0;
  
  ucg->xmega_hook = (void *) p;

}


// bitmap facilities

/*! \brief  Prints a bitmap to the display
 *
 *  \param  ucg      pointer to struct for the display
 *  \param  xoffset  the x-position of the upper left corner of the bitmap
 *  \param  yoffset  the y-position of the upper left corner of the bitmap
 *  \param  width    the width of the bitmap in pixels
 *  \param  height   the height of the bitmap in pixels
 *  \param  nbytes   the number of bytes of one pixel; must always be 3 (RGB)
 *  \param  bitmap   the pointer to a const unit8_t array with the bitmap.
 *
 *                   The bitmap must be smaller than 32K (largest AVR variable)
 *                   Large bitmaps can be split in smaller parts
 *                   The bitmaps can be located in the program memory or in the
 *                   data memory:
 *                   -# 'const uint8_t array[]' will be placed the data memory.
 *                          for atXmega256a3u devices the largest array ia 16K.
 *                   -# 'const uint8_t __flash array[]' will be placed only in the program memory.
 *                          for atXmega256a3u devices the maximum space is 64K
 *                   -# 'const uint8_t array[] PROGMEM' will be placed only in the program memory.
 *                          for atXmega256a3u devices the maximum space is 64K 
 *                   -# 'const uint8_t __memx array[]' will be placed only in the program memory.
 *                          with __memx the complete program space can be used 
 *                          for atXmega256a3u devices the maximum space is 256K
 *
 *                  The drawback of using __memx is that it is slower than __flash. This is
 *                  because _memx uses a 24-bits pointer and __flash uses a 16-bit pointer.
 *                  A sequence of two 30240 bytes bitmaps takes with __flash 3013 ms 
 *                  and with __memx 3063 ms.
 *
 *  \return void
 */
void  ucg_BitmapPrint(ucg_t *ucg, ucg_int_t xoffset,  ucg_int_t yoffset,
                                  ucg_int_t width, ucg_int_t height,
                                  uint8_t nbytes, const __memx uint8_t *bitmap)
{
  ucg_int_t  x = 0;
  ucg_int_t  y = 0;
  ucg_color_t oldColor;

  const __memx uint8_t *endBitmap = bitmap + width*height*nbytes;

  // remember original color
  memcpy(oldColor.color, ucg->arg.rgb[0].color, 3);

  while ( bitmap < endBitmap ) {
    ucg_SetColor(ucg, 0, *bitmap, *(bitmap+1), *(bitmap+2));
    bitmap += nbytes;
    ucg_DrawPixel(ucg, x+xoffset, y+yoffset);
    x++;
    if ( x >= width ) {
      x = 0;
      y++;
      if ( y >= height) break;
    }
  }

  // reset original color
  memcpy(ucg->arg.rgb[0].color, oldColor.color, 3);
}

/////////////////////////////////////////////////


// local functions communication 

/*  brief  Initialization of the communication 
 *
 *  return void
 */
static void _xmega_init(void)
{
  _commInterface.pRST->DIRSET = _commInterface.bmRST;
  _commInterface.pCD->DIRSET  = _commInterface.bmCD;
  if ( _commInterface.pBLK != NULL ) {     // BLK connected
    if ( _commInterface.blkDisabled ) {    // not used 
      *( (register8_t *) ( &(_commInterface.pBLK->PIN0CTRL) + _commInterface.bpBLK) ) = PORT_ISC_INPUT_DISABLE_gc;
    } else {                               // used
      _commInterface.pBLK->DIRSET = _commInterface.bmBLK;
      _commInterface.pBLK->OUTSET = _commInterface.bmBLK;
    }
  }      
  _commInterface.pSCK->DIRSET   = _commInterface.bmSCK;
  _commInterface.pSDA->DIRSET   = _commInterface.bmSDA;
  _commInterface.pCS->DIRSET    = _commInterface.bmCS;
  _commInterface.pCS->OUTSET    = _commInterface.bmCS;

  if ( _commInterface.pSPI != NULL ) {
    _commInterface.pSDI->DIRCLR  = _commInterface.bmSDI;
    _commInterface.pSPI->CTRL    = SPI_ENABLE_bm |  // enable SPI
                                   SPI_MASTER_bm |  // master mode
                                   SPI_CLK2X_bm  |  // double clock speed
                                // SPI_DORD_bm   |  // MSB first
                                   SPI_MODE_0_gc |  // SPI mode 0
                                   SPI_PRESCALER_DIV4_gc;  // prescaling 4
  }
}


/*  brief  Disable communication
*
*   return void
*/
static void _xmega_disable(void)
{
  if ( _commInterface.pSPI != NULL ) {
    _commInterface.pSPI->CTRL = _commInterface.pSPI->CTRL & ~SPI_ENABLE_bm;
  }
  if ( (_commInterface.pBLK != NULL) && (! _commInterface.blkDisabled) ) {   // connected and used
    _commInterface.pBLK->OUTCLR = _commInterface.bmBLK;
  }
}


/*  brief  Transfer a byte with bit banging
 *
 *  return void
 */
static inline void _xmega_transfer_bb(uint8_t data)
{
  for (uint8_t i=0; i<8; i++) {
    if (data & 0x80) {
      _commInterface.pSDA->OUTSET = _commInterface.bmSDA; 
    } else {
      _commInterface.pSDA->OUTCLR = _commInterface.bmSDA;
    }
    _commInterface.pSDA->OUTSET = _commInterface.bmSCK;
    _commInterface.pSDA->OUTCLR = _commInterface.bmSCK;
    data <<= 1;
  }
    
  _commInterface.pSDA->OUTSET = _commInterface.bmSDA;
}


/*  brief  Transfer a byte with SPI
 *
 *  return void
 */
static inline void _xmega_transfer_spi(uint8_t data)
{
  _commInterface.pSPI->DATA = data;
  while(!(_commInterface.pSPI->STATUS & (SPI_IF_bm)));
}

// local functions connection display Xmega

/*   brief  Gets (debugging) the last character of the PORT name
 *
 *   param  p   pointer to the PORT
 *
 *   return last character of the PORT name
 */
static char _get_port(PORT_t *p)
{
  switch ( (uint16_t) p ) {
    #ifdef PORTA
    case (uint16_t)&PORTA:  return 'A';
    #endif
    #ifdef PORTB
    case (uint16_t)&PORTB:  return 'B';
    #endif
    #ifdef PORTC
    case (uint16_t)&PORTC:  return 'C';
    #endif
    #ifdef PORTD
    case (uint16_t)&PORTD:  return 'D';
    #endif
    #ifdef PORTE
    case (uint16_t)&PORTE:  return 'E';
    #endif
    #ifdef PORTF
    case (uint16_t)&PORTF:  return 'F';
    #endif
    #ifdef PORTG
    case (uint16_t)&PORTG:  return 'G';
    #endif
  }

  return 'x';
}

/*  brief  Print (debugging) the SPI
 *
 *  param  s   pointer to the SPI
 *
 *  return void
 */
static char _get_spi(SPI_t *s)
{
  switch ( (uint16_t) s ) {
    #ifdef SPIC
    case (uint16_t)&SPIC:  return 'C';
    #endif
    #ifdef SPID
    case (uint16_t)&SPID:  return 'D';
    #endif
    #ifdef SPIE
    case (uint16_t)&SPIE:  return 'E';
    #endif
    #ifdef SPIF
    case (uint16_t)&SPIF:  return 'F';
    #endif
    #ifdef SPIG
    case (uint16_t)&SPIG:  return 'G';
    #endif
  }

  return 'x';
}

/*  brief  Print (debugging) the port 
 *
 *  param  c   last character of the PORT name
 *
 *  return void
 */
static void _print_port(char c)
{
  if ( c == 'x' ) {
    printf("%7s", "");
  } else {
    printf("PORT%c%2s", c, "");
  }
}

/*  brief  Print (debugging) the bitposition 
 *
 *  param  bp   bitposition
 *
 *  return void
 */
static void _print_bp(uint8_t bp)
{
  if ( bp > 7 ) {
    printf("%7s", "");
  } else {
    printf("%-6d%1s", bp, "");
  }
}

/*  brief  Print (debugging) the bitmask
 *
 *  param  bm   bitmask
 *
 *  return void
 */
static void _print_bm(uint8_t bm)
{
  if ( bm == 0 ) {
    printf("%7s", "");
  } else {
    printf("0x%02x%3s", bm, "");
  }
}

// local print facilities
/*  brief  Put a character to the display at the
 *         current position and in the current direction.
 *
 *  param  ucg      pointer to struct for the display
 *  param  dir      the direction
 *
 *  return void
 */
static int _ucg_putc(char c, FILE *stream)
{
  ucg_int_t delta;
  ucg_print_t *p = (ucg_print_t *) (_curr_ucg->xmega_hook);
  
  delta = ucg_DrawGlyph(_curr_ucg, p->tx, p->ty, p->tdir, c);
  
  switch(p->tdir) {
    case          0: p->tx += delta; break;
    case          1: p->ty += delta; break;
    case          2: p->tx -= delta; break;
    default: case 3: p->ty -= delta; break;
  }
  
  return 1;
}